"""Provides decorators to deal with tail calls in recursive functions."""
import builtins
import warnings
from collections import namedtuple
from Aspidites._vendor.contracts import contract
from Aspidites.api import create_warning


class TailCallWarning(RuntimeWarning):
    pass


class tco(object):
    """Provides a trampoline for functions that need one.

    Such function should return one of:
     * (False, result) - will exit from loop and return result to caller
     * (True, args) or (True, args, kwargs) - will repeat loop with the same
                                              function and other arguments
     * (func, args) or (func, kwargs) - will repeat loop with new callable
                                        and new arguments

    Usage example::

        @recur.tco
        def accumulate(origin, f=operator.add, acc=0):
            n = next(origin, None)
            if n is None: return False, acc
            return True, (origin, f, f(acc, n))

    Idea was described on python mailing list:
    http://mail.python.org/pipermail/python-ideas/2009-May/004486.html
    """

    __slots__ = "func",

    @contract()
    def __init__(self, func: 'Callable'):
        self.func = func

    def __delete__(self, instance):
        del self.func

    def __call__(self, *args, **kwargs):
        __self = self
        while True:
            result = __self.func(*args, **kwargs)
            # return final result
            if not result[0]:
                return result[1]
            # next loop with other arguments
            action, args = result[:2]
            # it's possible to given other function to run
            # XXX: do I need to raise exception if it's
            # impossible to use such function in tail calls loop?
            if callable(action):
                __self = action
            else:
                str_args = ''
                for i in args:
                    if hasattr(i, '__name__'):
                        str_args += i.__name__ + ', '
                    else:
                        str_args += str(type(i)) + ', '
                str_args = str_args.rstrip(', ')
                # DO NOT try inspecting the stack from here
                w = "\nOptimized tail call to %s(%s) failed, object of type '%s' is not callable.\n" % (str(action), str_args, type(action).__name__)
                warnings.warn(w, TailCallWarning)
            kwargs = result[2] if len(result) > 2 else {}


class stackless(object):
    """Provides a "stackless" (constant Python stack space) recursion
    decorator for generators.

    Invoking as f() creates the control structures. Within a
    function, only use `yield f.call()` and `yield f.tailcall()`.

    Usage examples:

    Tail call optimised recursion with tailcall()::

        @recur.stackless
        def fact(n, acc=1):
            if n == 0:
                yield acc
                return
            yield fact.tailcall(n-1, n*acc)

    Non-tail recursion with call() uses heap space so won't overflow::

        @recur.stackless
        def fib(n):
            if n == 0:
                yield 1
                return
            if n == 1:
                yield 1
                return
            yield (yield fib.call(n-1)) + (yield fib.call(n-2))

    Mutual recursion also works::

        @recur.stackless
        def is_odd(n):
            if n == 0:
                yield False
                return
            yield is_even.tailcall(n-1)

        @recur.stackless
        def is_even(n):
            if n == 0:
                yield True
                return
            yield is_odd.tailcall(n-1)

    """

    __slots__ = "func",

    Thunk = namedtuple("Thunk", ("func", "args", "kwargs", "is_tailcall"))

    def __init__(self, func):
        self.func = func

    def call(self, *args, **kwargs):
        return self.Thunk(self.func, args, kwargs, False)

    def tailcall(self, *args, **kwargs):
        return self.Thunk(self.func, args, kwargs, True)

    def __call__(self, *args, **kwargs):
        s = [self.func(*args, **kwargs)]
        r = []
        v = None
        while s:
            try:
                if r:
                    v = s[-1].send(r[-1])
                    r.pop()
                else:
                    v = next(s[-1])
            except StopIteration:
                s.pop()
                continue

            if isinstance(v, self.Thunk):
                g = v.func(*v.args, **v.kwargs)
                if v.is_tailcall:
                    s[-1] = g
                else:
                    s.append(g)
            else:
                r.append(v)
        return r[0] if r else None



from sys import version_info

if version_info[0] == 3:
    from functools import reduce

reduce = reduce

from itertools import filterfalse
from itertools import zip_longest

# Using or importing the ABCs from 'collections' instead of from
# 'collections.abc' is deprecated, and in 3.8 it will stop working.
if version_info[0] <= 3 and version_info[1] < 8:
    from collections import Iterable
else:
    from collections.abc import Iterable



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

import math
import random
import string


utility_builtins = {}

utility_builtins['string'] = string
utility_builtins['math'] = math
utility_builtins['random'] = random
utility_builtins['whrandom'] = random
utility_builtins['set'] = set
utility_builtins['frozenset'] = frozenset

try:
    import DateTime
    utility_builtins['DateTime'] = DateTime.DateTime  # pragma: no cover
except ImportError:
    pass


def same_type(arg1, *args):
    """Compares the class or type of two or more objects."""
    t = getattr(arg1, '__class__', type(arg1))
    for arg in args:
        if getattr(arg, '__class__', type(arg)) is not t:
            return False
    return True


utility_builtins['same_type'] = same_type


def test(*args):
    length = len(args)
    for i in range(1, length, 2):
        if args[i - 1]:
            return args[i]

    if length % 2:
        return args[-1]


utility_builtins['test'] = test


def reorder(s, with_=None, without=()):
    # s, with_, and without are sequences treated as sets.
    # The result is subtract(intersect(s, with_), without),
    # unless with_ is None, in which case it is subtract(s, without).
    if with_ is None:
        with_ = s
    orig = {}
    for item in s:
        if isinstance(item, tuple) and len(item) == 2:
            key, value = item
        else:
            key = value = item
        orig[key] = value

    result = []

    for item in without:
        if isinstance(item, tuple) and len(item) == 2:
            key, ignored = item
        else:
            key = item
        if key in orig:
            del orig[key]

    for item in with_:
        if isinstance(item, tuple) and len(item) == 2:
            key, ignored = item
        else:
            key = item
        if key in orig:
            result.append((key, orig[key]))
            del orig[key]

    return result


utility_builtins['reorder'] = reorder




# Aspidites
# Copyright (C) 2021 Ross J. Duff

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import py_compile
import sys
from glob import glob
from warnings import warn
import typing as t
from mypy import api
from .templates import lib, makefile, pyproject, setup
from pyrsistent import pmap
from hashlib import sha256
from pathlib import Path
from .final import final
from ._vendor.semantic_version import Version
from pyparsing import ParseResults


@final()
class CheckedFileStack:

    """A convenience class for reading file data streams to stdout or to checksum"""

    def __init__(self, initial=None, pre_size=128):
        if initial is None:
            initial = {}
        self._files = pmap(initial, pre_size)
        self.all_files = self._files.evolver()
        self.pre_size = pre_size

    def read(self, data, print__=False, hash_func=None):
        curr_hash = hash_func() if hash_func else {}
        chunk = data.read(self.pre_size)
        while chunk:
            curr_hash.update(chunk) if hash_func else {}
            chunk = data.read(self.pre_size)
            if print__:
                print(chunk, '')
        return curr_hash if hash_func else None

    def __checksum(self, fname, write=True, check=False):
        fname = Path(fname)
        base, name = fname.parent, fname.name
        base = Path(base)
        fname_sha256 = base/("." + name + ".sha256")

        if write:
            with open(fname, "rb") as data:
                curr_hash = self.read(data, hash_func=sha256)
                with open(fname_sha256, "wb") as digest:
                    digest.write(curr_hash.digest())
                return pmap({curr_hash.digest(): fname}).items()[0]  # immutable
        if check:
            with open(fname_sha256, "rb") as digest:
                with open(fname, "rb") as data:
                    curr_hash = self.read(data, hash_func=sha256)
                    old = digest.read()
                    new = curr_hash.digest()
                    if new == old:
                        print(
                            "sha256 digest check successful: %s, %s == %s"
                            % (fname, new.hex(), old.hex())
                        )
                        return new
                    else:
                        print(
                            "sha256 digest failure: %s, %s != %s"
                            % (fname, new.hex(), old.hex())
                        )
                        return ""

    def register(self, fname):
        self.all_files.set(*self.__checksum(fname))

    def commit(self):
        return self.all_files.persistent()

    def create_file(self, fname, mode, root='',
                    text="# THIS FILE IS GENERATED - DO NOT EDIT #") -> None:

        if len(str(root)) > 0:
            root = Path(root)
            file = root/fname
        else:
            file = fname
        try:
            open(file, mode).write(text)
        except FileExistsError:
            self.register(file)

    def finalize(self):
        all_file_checksums = self.commit()
        print("running checksums")
        for k, v in all_file_checksums.items():
            digest = self.__checksum(v, write=False, check=True)
            try:
                all_file_checksums.get(digest)
            except AttributeError:
                raise RuntimeError("\nfor file %s\n%s\n  did not match cached digest\n%s")


file_stack = CheckedFileStack()


def compile_module(**kwargs):
    code          : ParseResults         = kwargs['code']
    fname         : Path                 = kwargs['fname']
    force         : bool                 = kwargs['force']
    bytecode      : bool                 = kwargs['bytecode']
    c             : bool                 = kwargs['c']
    build_requires: t.Union[t.List, str] = kwargs['build_requires']
    verbose       : int                  = kwargs['verbose']

    fname = Path(fname)
    app_name = fname.parent / fname.stem
    project = app_name.stem
    module_name = str(app_name).replace("/", ".")
    file_c = str(app_name) + ".c"
    root = fname.parent
    mode = "x" if force else "w"

    files = {
        fname           : (mode, {'root': '', 'text': lib.substitute(code="\n".join(code))}),
        '__init__.py'   : (mode, {'root': root}),
        'py.typed'      : (mode, {'root': root}),
        'pyproject.toml': (mode,
            {'root': root, 'text': pyproject.substitute(build_requires=build_requires)}
                           ),
        'Makefile'      : (mode, {'root': root, 'text': makefile.substitute(project=project)}),
    }

    for k, v in files.items():
        args_, kwargs_ = v
        file_stack.create_file(k, args_, **kwargs_)

    typecheck(module_name)
    create_stubs(fname, app_name)

    if bytecode:
        fname_pyc = str(app_name) + ".pyc"
        quiet = tuple(reversed(range(3))).index(verbose if verbose < 2 else 2)
        major, minor, patch, *_ = sys.version_info
        if Version(major=major, minor=minor, patch=patch) < Version('3.8.0'):
            py_compile.compile(str(fname), fname_pyc)
        else:
            py_compile.compile(str(fname), fname_pyc, quiet=quiet)
        file_stack.register(fname_pyc)

    if c:
        compile_c(fname, force, verbose)
        create_setup(root, app_name, **kwargs)
        compile_object(str(app_name), file_c, root)

    file_stack.finalize()


def typecheck(module_name: str):
    mypy_args = [
        "-m",
        module_name,
        "--follow-imports=skip",
        "--show-error-context",
        "--show-error-codes",
        "--allow-incomplete-defs",
        "--disable-error-code=valid-type",
        "--exclude=builtins"
        # '--disallow-untyped-defs',
        # '--disallow-untyped-calls',
    ]
    print("running mypy", " ".join(mypy_args))
    type_report = None
    error_report = None
    return_code = 0
    try:
        type_report, error_report, return_code = api.run(mypy_args)
    except AttributeError:
        warn("mypy api call failed")

    finally:
        print("mypy type report: ", type_report) if type_report else None
        print("mypy error report: ", error_report) if error_report else None
        print("mypy returned with exit code:", return_code) if return_code else None
        # exit(return_code) if return_code != 0 else print("running compile")


def create_stubs(fname, app_name):
    fname_pyi = str(app_name) + '.pyi'
    stubgen = 'stubgen %s -o .' % fname
    print("running %s" % stubgen)
    with os.popen(stubgen) as p:
        file_stack.read(p, print__=True)
    try:
        file_stack.register(fname_pyi)
    except FileNotFoundError as e:
        warn(str(e))
        try:
            cwd = Path.cwd()
            path = cwd / Path('__main__.py')
            print("trying rename %s/__main__.pyi to %s" % (str(cwd), fname_pyi))
            path.rename(fname_pyi)
            file_stack.register(fname_pyi)
        except FileNotFoundError:
            warn("failed to create stubs")


def compile_c(fname, force, verbose) -> None:
    verb = int(bool(verbose))
    os.popen("cython %s %s %s" % (fname, "--force" * force, "--verbose" * verb))


def create_setup(root, app_name, **kwargs) -> None:
    module_name = str(app_name).replace("/", ".")
    file_stack.create_file('setup.py',
                "x" if kwargs['force'] else "w",
                root=root,
                text=setup.substitute(
                    app_name=module_name,
                    src_file=kwargs['fname'],
                    inc_dirs=[],
                    libs=[],
                    exe_name=app_name,
                    lib_dirs=[],
                    **kwargs))


def compile_object(app_name, file_c, root) -> None:
    glob_so = app_name + ".*.so"
    # TODO: get this working for docker builds
    #  (maybe executable param with os.path.relpath?)
    setup_runner = "%s %s build_ext -b ." % (sys.executable, str(Path(root)/'setup.py'))
    print("running", setup_runner)
    with os.popen(setup_runner) as p:
        file_stack.read(p, print__=True)
    file_stack.register(file_c)
    for i in glob(glob_so):
        file_stack.register(i)



import dataclasses
import warnings

from . import logger
import os


@dataclasses.dataclass(frozen=True)
class Switches:
    # ALWAYS ON
    disable_all = False


def disable_all():  # pragma: nocover
    """ Disables all contracts checks. """
#     print('disable_all()')
    warnings.warn('Attempted to call disable_all.', RuntimeWarning)


def enable_all():
    """
    Enables all contracts checks.
    Can be overridden by an environment variable.
    """
#     print('enable_all()')
    Switches.disable_all = False
    logger.info('All contracts checking enabled.')


def all_disabled():
    """ Returns true if all contracts checks are disabled. """
    return Switches.disable_all




# TODO: handle file mode?

import io
import sys

from ..interface import Contract, ContractNotRespected
from ..syntax import (add_contract, add_keyword, Keyword, W)

inPy2 = sys.version_info[0] == 2
if inPy2:  # pragma: no cover
    file_type = (file, io.IOBase)
else:
    file_type = io.IOBase


class File(Contract):

    def __init__(self, where=None):
        Contract.__init__(self, where)

    def check_contract(self, context, value, silent):

        if not isinstance(value, file_type):
            error = 'Expected a file, got %r.' % value.__class__.__name__
            raise ContractNotRespected(contract=self, error=error,
                                       value=value, context=context)

    def __str__(self):
        return 'file'

    def __repr__(self):
        return 'File()'

    @staticmethod
    def parse_action(s, loc, _):
        where = W(s, loc)
        return File(where=where)


file_contract = Keyword('file')

file_contract.setParseAction(File.parse_action)

add_contract(file_contract)
add_keyword('file')



from Aspidites._vendor.RestrictedPython import compile_restricted_exec
from Aspidites._vendor._compat import IS_PY3
from Aspidites.tests.helper import restricted_exec

import pytest


GLOBAL_EXAMPLE = """
def x():
    global a
    a = 11
x()
"""


def test_Global():
    glb = {'a': None}
    restricted_exec(GLOBAL_EXAMPLE, glb)
    assert glb['a'] == 11


# Example from:
# https://www.smallsurething.com/a-quick-guide-to-nonlocal-in-python-3/
NONLOCAL_EXAMPLE = """
def outside():
    msg = "Outside!"
    def inside():
        nonlocal msg
        msg = "Inside!"
        print(msg)
    inside()
    print(msg)
outside()
"""


@pytest.mark.skipif(
    not IS_PY3,
    reason="The `nonlocal` statement was introduced in Python 3.0.")
def test_Nonlocal():
    result = compile_restricted_exec(NONLOCAL_EXAMPLE)
    assert result.errors == ('Line 5: Nonlocal statements are not allowed.',)
    assert result.code is None



from itertools import starmap
from sys import version_info
from typing import Callable

from .func import F, identity
from .uniform import reduce


def _apply(f, args=None, kwargs=None):
    return f(*(args or []), **(kwargs or {}))


apply: Callable = _apply


def call(f, *args, **kwargs):
    return f(*args, **kwargs)


def flip(f):
    """Return function that will apply arguments in reverse order"""

    # Original function is saved in special attribute
    # in order to optimize operation of "duble flipping",
    # so flip(flip(A)) is A
    # Do not use this approach for underscore callable,
    # see https://github.com/kachayev/fn.py/issues/23
    flipper = getattr(f, "__flipback__", None)
    if flipper is not None:
        return flipper

    def _flipper(a, b):
        return f(b, a)

    setattr(_flipper, "__flipback__", f)
    return _flipper


def curry(f, arg, *rest):
    return curry(f(arg), *rest) if rest else f(arg)


def zipwith(f):
    """
    zipwith(f)(seq1, seq2, ..) ->
        [f(seq1[0], seq2[0], ..), f(seq1[1], seq2[1], ..), ...]
    """
    return F(starmap, f) << zip


def foldl(f, init=None):
    """Return function to fold iterator to scala value
    using passed function as reducer.

    Usage:
        >>> print foldl(_ + _)([0,1,2,3,4])
        10
        >>> print foldl(_ * _, 1)([1,2,3])
        6
    """

    def fold(it):
        args = [f, it]
        if init is not None:
            args.append(init)
        return reduce(*args)

    return fold


def foldr(f, init=None):
    """Return function to fold iterator to scala value using
    passed function as reducer in reverse order (consume values
    from iterator from right-to-left).

    Usage:
        >>> print foldr(call, 10)([lambda s: s**2, lambda k: k+10])
        400
    """

    def fold(it):
        args = [flip(f), reversed(it)]
        if init is not None:
            args.append(init)
        return reduce(*args)

    return fold


def unfold(f):
    """Return function to unfold value into stream using
    passed function as values producer. Passed function should
    accept current cursor and should return:

      * tuple of two elements (value, cursor), value will be added
        to output, cursor will be used for next function call
      * None in order to stop producing sequence

    Usage:
        >>> doubler = unfold(lambda x: (x*2, x*2))
        >>> list(islice(doubler(10), 0, 10))
        [20, 40, 80, 160, 320, 640, 1280, 2560, 5120, 10240]
    """

    def _unfolder(start):
        value, curr = None, start
        while 1:
            step = f(curr)
            if step is None:
                break
            value, curr = step
            yield value
    return _unfolder



import operator
import sys
from datetime import timedelta as delta
import hypothesis
import pytest as pt
import hypothesis.strategies as st
from hypothesis import given, assume

from Aspidites._vendor.fn import recur


# no fuzzing just tested at 10x the recursion limit
def test_tco_decorator(limit=sys.getrecursionlimit() * 10):

    def recur_accumulate(origin, f=operator.add, acc=0):
        n = next(origin, None)
        if n is None:
            return acc
        return recur_accumulate(origin, f, f(acc, n))

    # this works normally
    assert 10 == recur_accumulate(iter(range(5)))

    # such count of recursive calls should fail on CPython,
    # for PyPy we skip this test cause on PyPy the limit is
    # approximative and checked at a lower level
    # if not hasattr(sys, 'pypy_version_info'):
    #     with pt.raises(RuntimeError):
    #         recur_accumulate(iter(range(limit)))

    # with recur decorator it should run without problems
    @recur.tco
    def tco_accumulate(origin, f=operator.add, acc=0):
        n = next(origin, None)
        if n is None:
            return False, acc
        return True, (origin, f, f(acc, n))

    assert sum(range(limit)) == tco_accumulate(iter(range(limit)))


@hypothesis.settings(deadline=delta(milliseconds=400))
@given(st.integers(min_value=2, max_value=1000))
def test_tco_different_functions(i: int):
    # noinspection PyUnusedLocal
    assume(not i % 2)  # even only

    @recur.tco
    def recur_inc2(curr, acc=0):
        if curr == 0:
            return False, acc
        return recur_dec, (curr-1, acc+2)

    @recur.tco
    def recur_dec(curr, acc=0):
        if curr == 0:
            return False, acc
        return recur_inc2, (curr-1, acc-1)

    assert i/2 == recur_inc2(i)


def test_stackless():
    @recur.stackless
    def fib(n):
        if n == 0:
            yield 1
            return
        if n == 1:
            yield 1
            return
        yield (yield fib.call(n - 1)) + (yield fib.call(n - 2))
    assert fib(9) == 55



# cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8
# THIS FILE IS GENERATED - DO NOT EDIT #
import cython  # type: ignore
from typing import Any
from collections.abc import Generator
from pyrsistent import (
    pset, 
    pmap, 
    pvector, 
    s, v, m, 
    PRecord,
    PClass
)
from Aspidites.woma import *
from Aspidites._vendor import (
    reduce,
    filterfalse,
    zip_longest,
    accumulate,
    take,
    drop,
    takelast,
    droplast,
    _,
)
from Aspidites.monads import Maybe, Surely, Undefined, SafeDiv, SafeMod, SafeExp
from Aspidites._vendor.contracts import contract, new_contract
from Aspidites._vendor.RestrictedPython import safe_builtins
safe_builtins['print'] = print
from Aspidites._vendor.RestrictedPython import compile_restricted as compile
safe_builtins['compile'] = compile
# DECLARATIONS TO ALLOW CONTRACTS TO TYPE CHECK #
procedure: None
coroutine: Generator
number: Any
globals().update(dict(__builtins__=safe_builtins))  # add all imports to globals


@contract()
@cython.binding(True)
def Add(x : 'number' = 0, y : 'number' = 0) -> 'number':
    return x+y


@contract()
@cython.binding(True)
def Sub(x : 'number' = 0, y : 'number' = 0) -> 'number':
    return x-y


@contract()
@cython.binding(True)
def Div(x : 'number' = 0, y : 'number' = 0) -> 'number':
    return Maybe(SafeDiv, x, y)


@contract()
@cython.binding(True)
def Exp(x : 'number' = 0, y : 'number' = 0) -> 'number':
    return Maybe(SafeExp, x, y)


@contract()
@cython.binding(True)
def Mod(x : 'number' = 0, y : 'number' = 0) -> 'number':
    return Maybe(SafeMod, x, y)






"""
Tests about imports
"""

from Aspidites._vendor.RestrictedPython import compile_restricted_exec
from Aspidites._vendor.RestrictedPython import safe_builtins
from Aspidites._vendor._compat import IS_PY2
from Aspidites._vendor._compat import IS_PY3
from Aspidites.tests.helper import restricted_exec

import pytest


OS_IMPORT_EXAMPLE = """
import os

os.listdir('/')
"""


def test_os_import():
    """It does not allow to import anything by default.

    The `__import__` function is not provided as it is not safe.
    """
    # Caution: This test is broken on PyPy until the following issue is fixed:
    # https://bitbucket.org/pypy/pypy/issues/2653
    # PyPy currently ignores the restriction of the `__builtins__`.
    glb = {'__builtins__': safe_builtins}

    with pytest.raises(ImportError) as err:
        restricted_exec(OS_IMPORT_EXAMPLE, glb)
    assert '__import__ not found' == str(err.value)


__BUILTINS_EXAMPLE = """
import __builtins__

mygetattr = __builtins__['getattr']
"""


@pytest.mark.skipif(
    IS_PY3,
    reason="__builtins__ has been renamed in Python3 to builtins, "
    "so no __builtins__ in Python3.")
def test_import_py2_builtins():  # pragma: no cover
    """It should not be allowed to access global __builtins__ in Python2."""
    result = compile_restricted_exec(__BUILTINS_EXAMPLE)
    assert result.code is None
    assert result.errors == (
        'Line 2: "__builtins__" is an invalid variable name because it starts with "_"',  # NOQA: E501
        'Line 4: "__builtins__" is an invalid variable name because it starts with "_"'  # NOQA: E501
    )


__BUILTINS_AS_EXAMPLE = """
import __builtins__ as builtins

mygetattr = builtins['getattr']
"""


@pytest.mark.skipif(
    IS_PY3,
    reason="__builtins__ has been renamed in Python3 to builtins, "
    "so no __builtins__ in Python3.")
def test_import_py2_as_builtins():  # pragma: no cover
    """It should not be allowed to access global __builtins__ in Python2."""
    result = compile_restricted_exec(__BUILTINS_EXAMPLE)
    assert result.code is None
    assert result.errors == (
        'Line 2: "__builtins__" is an invalid variable name because it starts with "_"',  # NOQA: E501
        'Line 4: "__builtins__" is an invalid variable name because it starts with "_"'  # NOQA: E501
    )


BUILTINS_EXAMPLE = """
import builtins

mygetattr = builtins['getattr']
"""


@pytest.mark.skipif(
    IS_PY2,
    reason="__builtins__ has been renamed in Python3 to builtins, "
    "and need only to be tested there.")
def test_import_py3_builtins():  # pragma: PY3
    """It should not be allowed to access global builtins in Python3."""
    result = compile_restricted_exec(BUILTINS_EXAMPLE)
    assert result.code is None
    assert result.errors == (
        'Line 2: "builtins" is a reserved name.',
        'Line 4: "builtins" is a reserved name.'
    )


BUILTINS_AS_EXAMPLE = """
import builtins as glb

mygetattr = glb['getattr']
"""


@pytest.mark.skipif(
    IS_PY2,
    reason="__builtins__ has been renamed in Python3 to builtins, "
    "and need only to be tested there.")
def test_import_py3_as_builtins():  # pragma: PY3
    """It should not be allowed to access global builtins in Python3."""
    result = compile_restricted_exec(BUILTINS_AS_EXAMPLE)
    assert result.code is None
    assert result.errors == ('Line 2: "builtins" is a reserved name.',)



from Aspidites._vendor.RestrictedPython.Guards import guarded_iter_unpack_sequence
from Aspidites.tests.helper import restricted_exec

import types


ITERATORS = """
def for_loop(it):
    c = 0
    for a in it:
        c = c + a
    return c


def nested_for_loop(it1, it2):
    c = 0
    for a in it1:
        for b in it2:
            c = c + a + b
    return c

def dict_comp(it):
    return {a: a + a for a in it}

def list_comp(it):
    return [a + a for a in it]

def nested_list_comp(it1, it2):
    return [a + b for a in it1 if a > 1 for b in it2]

def set_comp(it):
    return {a + a for a in it}

def generator(it):
    return (a + a for a in it)

def nested_generator(it1, it2):
    return (a+b for a in it1 if a > 0 for b in it2)
"""


def test_RestrictingNodeTransformer__guard_iter__1(mocker):
    it = (1, 2, 3)
    _getiter_ = mocker.stub()
    _getiter_.side_effect = lambda x: x
    glb = {'_getiter_': _getiter_}
    restricted_exec(ITERATORS, glb)

    ret = glb['for_loop'](it)
    assert 6 == ret
    _getiter_.assert_called_once_with(it)
    _getiter_.reset_mock()

    ret = glb['nested_for_loop']((1, 2), (3, 4))
    assert 20 == ret
    _getiter_.assert_has_calls([
        mocker.call((1, 2)),
        mocker.call((3, 4))
    ])
    _getiter_.reset_mock()

    ret = glb['dict_comp'](it)
    assert {1: 2, 2: 4, 3: 6} == ret
    _getiter_.assert_called_once_with(it)
    _getiter_.reset_mock()

    ret = glb['list_comp'](it)
    assert [2, 4, 6] == ret
    _getiter_.assert_called_once_with(it)
    _getiter_.reset_mock()

    ret = glb['nested_list_comp']((1, 2), (3, 4))
    assert [5, 6] == ret
    _getiter_.assert_has_calls([
        mocker.call((1, 2)),
        mocker.call((3, 4))
    ])
    _getiter_.reset_mock()

    ret = glb['set_comp'](it)
    assert {2, 4, 6} == ret
    _getiter_.assert_called_once_with(it)
    _getiter_.reset_mock()

    ret = glb['generator'](it)
    assert isinstance(ret, types.GeneratorType)
    assert list(ret) == [2, 4, 6]
    _getiter_.assert_called_once_with(it)
    _getiter_.reset_mock()

    ret = glb['nested_generator']((0, 1, 2), (1, 2))
    assert isinstance(ret, types.GeneratorType)
    assert list(ret) == [2, 3, 3, 4]
    _getiter_.assert_has_calls([
        mocker.call((0, 1, 2)),
        mocker.call((1, 2)),
        mocker.call((1, 2))])
    _getiter_.reset_mock()


ITERATORS_WITH_UNPACK_SEQUENCE = """
def for_loop(it):
    c = 0
    for (a, b) in it:
        c = c + a + b
    return c

def dict_comp(it):
    return {a: a + b for (a, b) in it}

def list_comp(it):
    return [a + b for (a, b) in it]

def set_comp(it):
    return {a + b for (a, b) in it}

def generator(it):
    return (a + b for (a, b) in it)
"""


def test_RestrictingNodeTransformer__guard_iter__2(mocker):
    it = ((1, 2), (3, 4), (5, 6))

    call_ref = [
        mocker.call(it),
        mocker.call(it[0]),
        mocker.call(it[1]),
        mocker.call(it[2])
    ]

    _getiter_ = mocker.stub()
    _getiter_.side_effect = lambda x: x

    glb = {
        '_getiter_': _getiter_,
        '_iter_unpack_sequence_': guarded_iter_unpack_sequence
    }

    restricted_exec(ITERATORS_WITH_UNPACK_SEQUENCE, glb)

    ret = glb['for_loop'](it)
    assert ret == 21
    _getiter_.assert_has_calls(call_ref)
    _getiter_.reset_mock()

    ret = glb['dict_comp'](it)
    assert ret == {1: 3, 3: 7, 5: 11}
    _getiter_.assert_has_calls(call_ref)
    _getiter_.reset_mock()

    ret = glb['list_comp'](it)
    assert ret == [3, 7, 11]
    _getiter_.assert_has_calls(call_ref)
    _getiter_.reset_mock()

    ret = glb['set_comp'](it)
    assert ret == {3, 7, 11}
    _getiter_.assert_has_calls(call_ref)
    _getiter_.reset_mock()

    ret = list(glb['generator'](it))
    assert ret == [3, 7, 11]
    _getiter_.assert_has_calls(call_ref)
    _getiter_.reset_mock()



from Aspidites._vendor.RestrictedPython.Eval import RestrictionCapableEval as RestrictionCapableEval
from Aspidites._vendor.RestrictedPython.Guards import safe_builtins as safe_builtins, safe_globals as safe_globals
from Aspidites._vendor.RestrictedPython.Limits import limited_builtins as limited_builtins
from Aspidites._vendor.RestrictedPython.PrintCollector import PrintCollector as PrintCollector
from Aspidites._vendor.RestrictedPython.Utilities import utility_builtins as utility_builtins
from Aspidites._vendor.RestrictedPython.compile import CompileResult as CompileResult, compile_restricted as compile_restricted, compile_restricted_eval as compile_restricted_eval, compile_restricted_exec as compile_restricted_exec, compile_restricted_function as compile_restricted_function, compile_restricted_single as compile_restricted_single
from Aspidites._vendor.RestrictedPython.transformer import RestrictingNodeTransformer as RestrictingNodeTransformer
